/*
  This file is a part of hashutils package.
  Copyright (C) 2004 Zeljko Vrba.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  You can reach me at the following e-mail addresses:

  zvrba@globalnet.hr
  mordor@fly.srk.fer.hr
*/
#include <stddef.h>
#include "cdt_interval.h"

static char rcsid[] = "$Id: cdt_interval.c,v 1.3 2004/03/23 17:43:24 zvrba Exp $";

static Vmalloc_t *IntervalSet_vmpool = NULL;

/**
   Intervals X=[lb1, ub1) and Y=[lb2, ub2) are totally ordered by the following
   rules:
   1. X < Y  <=> ub1 < lb2
   2. X > Y  <=> lb1 > ub2
   3. otherwise they are considered 'equal' and can be merged in a single
      interval.
*/
static int interval_comparf(Dt_t *dt, Void_t *key1, Void_t *key2, Dtdisc_t *disc)
{
	unsigned long *i1 = key1, *i2 = key2;

	if(i1[1] < i2[0]) return -1;
	if(i1[0] > i2[1]) return 1;
	return 0;
}

static Void_t *interval_makef(Dt_t *dt, Void_t *obj, Dtdisc_t *disc)
{
	return vmalloc(IntervalSet_vmpool, sizeof(struct interval));
}

static void interval_freef(Dt_t *dt, Void_t *obj, Dtdisc_t *disc)
{
	vmfree(IntervalSet_vmpool, obj);
}

Dtdisc_t Cdt_IntervalDiscipline = {
	offsetof(struct interval, bounds), /* key offset */
	2*sizeof(unsigned long),	/* key size */
	0,							/* link: beginning of struct */
	interval_makef,			    /* copy method */
	interval_freef,				/* free method */
	interval_comparf,			/* compare */
	NULL,						/* hash: use internal function */
	NULL,
	NULL
};

Dt_t *IntervalSet_create(void)
{
	if(!IntervalSet_vmpool &&
	   !(IntervalSet_vmpool = vmopen(Vmdcheap, Vmpool, VM_TRUST))) return NULL;
	return dtopen(&Cdt_IntervalDiscipline, Dtobag);
}

/**
   Insert a single number in the set of intervals. This just inserts the
   interval [n, n+1).

   @param	n	Number to insert.

   @return Pointer to interval in which the number was inserted.
*/
struct interval *IntervalSet_insert1(Dt_t *dt, unsigned long n)
{
	return IntervalSet_insert2(dt, n, n+1);
}

/**
   Merge two intervals.

   @return Merged interval. On error, an empty interval [0,0) is returned.
*/
static struct interval interval_merge(
	const struct interval *i1,
	const struct interval *i2)
{
	struct interval i;

	/* swap: i1 lower bound should be <= to i2 lower bound. */
	if(i1->bounds[0] > i2->bounds[0]) {
		const struct interval *tmp = i1; i1 = i2; i2 = tmp;
	}

	/* the upper bound of resulting interval is greater of both upper bounds */
	if(i2->bounds[0] <= i1->bounds[1]) {
		i.bounds[0] = i1->bounds[0];
		i.bounds[1] = i1->bounds[1] > i2->bounds[1] ? 
			i1->bounds[1] : i2->bounds[1];
	} else {
		/* intervals don't even touch; return empty interval */
		i.bounds[0] = i.bounds[1] = 0;
	}
	return i;
}

/*
  Insert an interval [lb, ub) in the set of intervals.
  @param	lb,ub	Lower and upper bound of interval.
  
  @return	Pointer to resulting interval or NULL on failure.

  @note	First all intervals 'equal' to the new interval are sought for,
  deleted and merged with the new interval. Then a single new interval is
  inserted.
*/
struct interval *IntervalSet_insert2(
	Dt_t *dt, unsigned long lb, unsigned long ub)
{
	struct interval result, *found;

	result.bounds[0] = lb; result.bounds[1] = ub;

	while((found = dtmatch(dt, result.bounds))) {
		result = interval_merge(&result, found);
		if(!result.bounds[0] && !result.bounds[1]) return NULL;
		dtdelete(dt, found);
	}
	return dtinsert(dt, &result); /* discipline copies the struct */
}
