/*
  This file is a part of hashutils package.
  Copyright (C) 2004 Zeljko Vrba.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  You can reach me at the following e-mail addresses:

  zvrba@globalnet.hr
  mordor@fly.srk.fer.hr
*/
#include <stdlib.h>
#include <string.h>
#include <sfio.h>
#include <cdt.h>
#include <mycrypt.h>
#include "halfhash.h"
#include "crypthash.h"
#include "main.h"

static char rcsid[] = "$Id: main.c,v 1.6 2004/02/22 09:21:27 zvrba Exp $";

/**
   Selected hash and filenames. If filename starts with '$(', then it must
   be processed as a pipe.
   @see \e Main_open_file
*/
static char *Main_hash, *Main_fname[2] = { 0 };

static void register_ciphers(void)
{
	register_cipher(&blowfish_desc);
	register_cipher(&rc2_desc);
	register_cipher(&rc5_desc);
	register_cipher(&safer_k64_desc);
	register_cipher(&safer_sk64_desc);
	register_cipher(&cast5_desc);
}

static void register_hashes(void)
{
	register_hash(&md2_desc);
	register_hash(&md4_desc);
	register_hash(&md5_desc);
	register_hash(&rmd128_desc);

	register_hash(&md2half_desc);
	register_hash(&md4half_desc);
	register_hash(&md5half_desc);
	register_hash(&rmd128half_desc);

	register_hash(&dm_blowfish_desc);
	register_hash(&dm_rc2_desc);
	register_hash(&dm_rc5_desc);
	register_hash(&dm_safer_k64_desc);
	register_hash(&dm_safer_sk64_desc);
	register_hash(&dm_cast5_desc);
}

static char *parse_fname(int argc, char **argv, int *i)
{
	Sfio_t *str;
	int pipe_start = 0;
	
	if(strcmp(argv[*i], "$(")) {
		char *ret = argv[*i];
		++*i;
		return ret;
	}
	pipe_start = *i;

	if(!(str = sftmp(SF_UNBOUND))) {
		sfprintf(sfstderr, "ERROR: can't create temp file, exiting.\n");
		exit(1);
	}
	
	sfprintf(str, "$(");
	while(++*i < argc) {
		if(!strcmp(argv[*i], ")$")) {
			if(*i - pipe_start < 2) {
				sfprintf(sfstderr, "ERROR: $()$ requires at least the command.\n");
				exit(1);
			}
			++*i;				/* advance past )$ */
			sfseek(str, 0, 0);
			return sfreserve(str, -1, 0);
		}
		sfprintf(str, "%s ", argv[*i]);
	}
	sfprintf(sfstderr, "ERROR: unterminated $(\n");
	exit(1);
}

void Main_parse_cmdline(int argc, char **argv)
{
	int i;
	
	/* process options */
	for(i = 1; (i < argc) && (argv[i][0] == '-'); i++) {
		if(!Main_process_option(argc, argv, &i)) {
			Main_usage("ERROR: invalid option.");
		}
	}
	
	/* parse hash */
	if(i < argc) Main_hash = argv[i++];
	else Main_usage("ERROR: Missing HASH argument.");

	/* parse at least 1 filename */
	if(i < argc) Main_fname[0] = parse_fname(argc, argv, &i);
	else Main_usage("ERROR: at least one file must be specified.");

	if(i < argc) Main_fname[1] = parse_fname(argc, argv, &i);
}

void Main_print_hashes(void)
{
	int i, err;

	for(i = 0; hash_descriptor[i].name; i++) {
		sfprintf(sfstderr, "  %s/%u",
				 hash_descriptor[i].name,
				 hash_descriptor[i].hashsize);
		err = hash_descriptor[i].test();
		if(err == CRYPT_OK) {
			sfprintf(sfstderr, "\tself test OK");
		} else if(err == CRYPT_NOP) {
			sfprintf(sfstderr, "\tself test NOT IMPLEMENTED.");
		} else {
			sfprintf(sfstderr, "\tself test FAILED!");
		}
		sfprintf(sfstderr, "\n");
	}
	sfprintf(sfstderr, "Note that md2, md2hx and all DM* ciphers are EXTREMELY SLOW.\n");
}

/**
   If \e fname begins with $(, open it as a pipe. Report an error if open
   doesn't succeed.
*/
Sfio_t *Main_file_open(const char *fname)
{
	Sfio_t *ret;
	
	if(!strcmp(fname, "-")) {
		ret = sfstdin;
	} else if(strncmp(fname, "$(", 2)) {
		if(!(ret = sfopen(NULL, fname, "rt"))) {
			sfprintf(sfstderr, "%s: ERROR: can't open %s: ", Main_argv0, fname);
			perror("sfopen");
			exit(1);
		}
	} else if(!(ret = sfpopen(NULL, fname+2, "r"))) {
		sfprintf(sfstderr, "%s: ERROR: can't execute %s: ", Main_argv0, fname);
		perror("sfpopen");
		exit(1);
	}
	return ret;
}

int main(int argc, char **argv)
{
	int halg;

	Main_argv0 = argv[0];
	
	/* init crypt library */
	init_half_hashes();
	register_hashes();
	register_ciphers();

	Main_parse_cmdline(argc, argv);
	
	if((halg = find_hash(Main_hash)) < 0) {
		Main_usage("ERROR: invalid hash name.");
	}
	
	/* open files */
	if(Main_fname[1] && !strcmp(Main_fname[0], Main_fname[1])) {
		sfprintf(sfstderr, "%s: ERROR: FROM-FILE and TO-FILE are same.\n", argv[0]);
		return 1;
	}

	Main_process_files(Main_fname[0], Main_fname[1], halg);

	return 0;
}
